#ifndef QNZBCREATOR_H
#define QNZBCREATOR_H

#include <QObject>
#include <QDateTime>

class QNZBCreator : public QObject
{
    Q_OBJECT
public:
    explicit QNZBCreator(QObject *parent = 0);

    /**
     * Group:
     * @newsgroup:              The name of the newsgroup
     * @group_id:               An identifier for this group
     */
    struct Group {
        QString newsgroup;
        int group_id;
    };

    enum SearchOnlyIn {
        SEARCH_ONLY_IN_SUBJECTS,
        SEARCH_ONLY_IN_FILENAMES,
        SEARCH_ONLY_IN_POSTERS
    };

    enum FileType {
        FILE_TYPE_UNKNOWN,
        FILE_TYPE_RAR,
        FILE_TYPE_PAR,
        FILE_TYPE_PAR2,
        FILE_TYPE_SFV,
        FILE_TYPE_NFO,
        FILE_TYPE_NZB,
        FILE_TYPE_OTHER
    };

    /**
     * SearchOpts:
     * @query:                  The search query
     * @username:               The username which will be used to perform the search query (unused in NNTPGrab 0.7)
     * @password:               The password which will be used to perform the search query (unused in NNTPGrab 0.7)
     * @search_only_in:         Indicate what type of data needs to be searched
     * @max_age:                The maximum age of search results (in days)
     * @minimal_file_size:      The minimum size of files in bytes
     * @maximal_file_size:      The maximum size of files in bytes
     * @percentage_complete:    Only return results which are this percentage complete (0 to 100)
     * @group_to_search:        Limit searching to one specific groups (see the @group_id field from #NZBCreatorGroup)
     * @file_type:              Only search files of this type
     */
    struct SearchOpts {
        QString query;
        QString username;
        QString password;
        QNZBCreator::SearchOnlyIn search_only_in;
        int max_age;
        qint64 minimal_file_size;
        qint64 maximal_file_size;
        int percentage_complete;
        int group_to_search;
        QNZBCreator::FileType file_type;
    };

    /**
     * File:
     * @file_id:                An identifier for this file
     * @subject:                The subject of this file
     * @poster:                 The poster of this file
     * @num_parts_found:        The number of parts which are found which belong to this file
     * @num_parts_expected:     The number of parts which are expected which belong to this file
     * @file_type:              The type of this file
     * @file_size:              The size of this file in bytes
     * @stamp:                  The stamp of the first part of this file
     * @complete_percentage:    The percentage which indicates how complete this file is
     */
    struct File {
        qint64 file_id;
        QString subject;
        QString poster;
        int num_parts_found;
        int num_parts_expected;
        QNZBCreator::FileType file_type;
        qint64 file_size;
        QDateTime stamp;
        int complete_percentage;
    };

    /**
     * Collection:
     * @collection_name:        The name of the collection (most frequently this is the subject of the first file in the collection)
     * @short_collection_name:  An abbreviated version of @collection_name
     * @newsgroup:              The newsgroup in which all the files from this collection are poster
     * @total_file_size:        The total size in bytes of this collection
     * @stamp:                  The stamp of the first part of the first file in this collection
     * @files:                  A list containing the files belonging to this collection
     */
     struct Collection {
        QString collection_name;
        QString short_collection_name;
        QString newsgroup;
        qint64 total_file_size;
        QDateTime stamp;
        QList<QNZBCreator::File> files;
    };

    /**
     * SearchResult:
     * @number_of_hits:         The number of hits which were found for the given query
     * @number_of_results:      The number of files which were found and are part of the @collections
     * @collections:            A list containing all returned collections
     */
    struct SearchResult {
        int number_of_hits;
        int number_of_results;
        QList<QNZBCreator::Collection> collections;
    };


    /**
     * checkApiVversion:
     *
     * Verify if this version of NNTPGrab can be used to communicate with the NZBCreator service
     *
     * Returns:      TRUE when this version of NNTPGrab can communicate with the NZBCreator service
     */
    bool checkApiVersion(void);

    /**
     * getAllGroups:
     * @errmsg:         If an errors occurs, the reason will be placed in this field
     * @groups:         Location where the list of groups can be saved
     *
     * Retrieve a list of all the usenet groups which are indexed by the NZBCreator service
     *
     * Returns:         TRUE on success, FALSE on failure (errmsg will be set)
     */
    bool getAllGroups(QString &errmsg, QList<QNZBCreator::Group> &groups);

    /**
     * performSearch:
     * @opts:     A structure containing information about the requested search
     * @errmsg:   If an errors occurs, the reason will be placed in this field
     * @result:   Location where the search results can be saved
     *
     * Perform a search using the NZBCreator service
     *
     * Returns:   TRUE on success, FALSE on failure (errmsg will be set)
     */
    bool performSearch(QNZBCreator::SearchOpts opts, QString &errmsg, QNZBCreator::SearchResult &result);

    /**
     * generateNZB:
     * @file_ids:  A list containing file ID's of all the files which needs to be retrieved
     * @errmsg:    If an errors occurs, the reason will be placed in this field
     *
     * Generate a NZB file based on a list of file ID's
     *
     * Returns:    The contents of the NZB file containing all the details of the given file_ids. If an error occurs, an empty string will be returned and errmsg will be set
     */
    QString generateNZB(QList<int> file_ids, QString &errmsg);

signals:

public slots:

};

#endif // QNZBCREATOR_H
