#include <QTranslator>
#include <QLocale>
#include <QLibraryInfo>
#include <QTimer>
#include "DialogConnect.h"
#include <QMessageBox>
#include <stdio.h>

#ifdef Q_OS_MAC
#include <CoreFoundation/CFBundle.h>
#endif

#ifdef WIN32
#include <winsock2.h>
#endif

static void messagehandler(QtMsgType type, const char *msg);

int main(int argc, char *argv[])
{
    qInstallMsgHandler(messagehandler);
    NGApplication a(argc, argv);

#ifdef WIN32
    /* Initialise Winsock */
    WORD wVersionRequested;
    WSADATA wsaData;
    int ret;

    wVersionRequested = MAKEWORD(2, 2);

    ret = WSAStartup(wVersionRequested, &wsaData);
    if (ret != 0) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        qWarning(QObject::tr("Winsock could not be initialised").toUtf8().data());
        return -1;
    }

    /* Confirm that the WinSock DLL supports 2.2.        */
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */

    if (LOBYTE( wsaData.wVersion ) != 2 ||
        HIBYTE( wsaData.wVersion ) != 2) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup();

        qWarning(QObject::tr("Winsock 2.2 or higher is required for this program").toUtf8().data());

        return -1;
    }

    // exchndl.dll is part of Dr. MinGW. By loading this DLL, every time the
    // program crashes a .RPT file will be created containing the backtrace
    LoadLibraryA("exchndl.dll");
#endif

    /* Verify that there's no other instance of NNTPGrab running already */
    if (a.sendMessage("wakeUp")) {
        /* There is another instance active. Check if we need to open a NZB file */
        bool beenhere = false;
        Q_FOREACH(QString arg, a.arguments()) {
            /* Ignore the first argument (which is the executable) */
            if (!beenhere) {
                beenhere = true;
                continue;
            }

            /* Test if the argument is a file name */
            QFile f(arg);
            if (f.exists()) {
                /* Request the other instance to open the file */
                a.sendMessage(arg);
            }
        }

        return 0;
    }

    if (a.arguments().contains("--debug", Qt::CaseInsensitive)) {
        qputenv("NNTPGRAB_DEBUG", "1");
    }

    QTranslator qtTranslator;
    qtTranslator.load("qt_" + QLocale::system().name(), QLibraryInfo::location(QLibraryInfo::TranslationsPath));
    a.installTranslator(&qtTranslator);

    QTranslator nntpgrabTranslator;
#ifdef Q_OS_LINUX
    /* Yes this is very ugly..but as QMake doesn't have a default path for translations on Linux environments, this is the best we have */
    nntpgrabTranslator.load("nntpgrab_" + QLocale::system().name(), QString("/usr/share/nntpgrab/translations"));
#elif defined(Q_OS_WIN32)
    nntpgrabTranslator.load("nntpgrab_" + QLocale::system().name(), QCoreApplication::applicationDirPath() + QString("/../share/nntpgrab/translations"));
#elif defined(Q_OS_MAC)
    CFURLRef appUrlRef = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle());
    CFStringRef macPath = CFURLCopyFileSystemPath(appUrlRef, kCFURLPOSIXPathStyle);
    QString resources_path = CFStringGetCStringPtr(macPath, CFStringGetSystemEncoding());
    CFRelease(appUrlRef);
    CFRelease(macPath);

    appUrlRef = CFBundleCopyBundleURL(CFBundleGetMainBundle());
    macPath = CFURLCopyFileSystemPath(appUrlRef, kCFURLPOSIXPathStyle);
    QString bundle_path = CFStringGetCStringPtr(macPath, CFStringGetSystemEncoding());
    CFRelease(appUrlRef);
    CFRelease(macPath);

    nntpgrabTranslator.load("nntpgrab_" + QLocale::system().name(), bundle_path + QString("/") + resources_path + QString("/share/nntpgrab/translations"));
#else
    nntpgrabTranslator.load("nntpgrab_" + QLocale::system().name());
#endif
    a.installTranslator(&nntpgrabTranslator);

    DialogConnect d(&a);
    ConfigGuiOpts gui_opts;
    if (!gui_opts.getShowConnectDialog() && !QNNTPGrabGlue::utilsTestIsServerAlreadyRunning()) {
        /* Start the standalone mode as soon as the main loop has been started */
        QTimer::singleShot(0, &d, SLOT(doStartStandalone()));
    } else {
        d.show();
    }
    return a.exec();
}

static void messagehandler(QtMsgType type, const char *msg)
{
    switch (type) {
    case QtDebugMsg:
        puts(msg);
        break;
    case QtWarningMsg:
    case QtCriticalMsg:
        QMessageBox::warning(NULL,
                             "NNTPGrab Warning",
                             msg,
                             QMessageBox::Ok);
        break;
    case QtFatalMsg:
        QMessageBox::critical(NULL,
                              "NNTPGrab Error",
                              msg,
                              QMessageBox::Ok);
        QCoreApplication::exit(1);
        break;
    }
}

