#include "mainwindow.h"
#include "ui_mainwindow.h"

#ifdef Q_WS_WIN
#include <windows.h>
#endif

#include <QDesktopServices>
#include <QUrl>

#include "DialogManagePlugins.h"
#include "DialogImportNZB.h"
#include "ConfigGuiOpts.h"

#pragma GCC diagnostic ignored "-Wwrite-strings"
#include "icon_64x64.xpm"
#pragma GCC diagnostic warning "-Wwrite-strings"

MainWindow::MainWindow(QNNTPGrabGlue *glue, NGApplication *app, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    this->glue = glue;
    this->formDebug = NULL;
    this->showNotificationOnFileDownloaded = false;
    this->showNotificationOnFileDecoded = false;

    ui->setupUi(this);

    /* Create the individual tabs */
    updateInfoWidget = new WidgetUpdateInfo();
    ui->tabWidget->addTab(updateInfoWidget, tr("Information"));

    /* Create the traffic monitor and list of active connections */
    trafficMonitorWidget = new WidgetTrafficMonitor(glue);
    ui->splitterBottom->addWidget(trafficMonitorWidget);
    connectionsWidget = new WidgetConnections(glue);
    ui->splitterBottom->addWidget(connectionsWidget);

    downloadQueueWidget = new WidgetDownloadQueue(glue, ui->actionRemoveSelectedFiles);

    configWidget = new WidgetConfig(glue, connectionsWidget, downloadQueueWidget, this);
    ui->tabWidget->addTab(configWidget, tr("Configuration"));

    ui->tabWidget->addTab(downloadQueueWidget, tr("Download queue"));

    nzbcreatorWidget = new WidgetNZBCreator(glue);
    ui->tabWidget->addTab(nzbcreatorWidget, tr("Online Search"));

    repairAndUnpackWidget = new WidgetRepairAndUnpack(glue);
    ui->tabWidget->addTab(repairAndUnpackWidget, tr("Repair and Unpack"));

    /* Remove the dummy tab */
    ui->tabWidget->removeTab(0);

    /* Fix the position of the splitter */
    /* FIXME: This must be implemented in a more proper way.. */
    QList<int> sizes = ui->splitterBottom->sizes();
    sizes[0] = 150;
    sizes[1] = 65535;
    ui->splitterBottom->setSizes(sizes);

    sizes = ui->splitterMain->sizes();
    sizes[0] = 65535;
    sizes[1] = 150;
    ui->splitterMain->setSizes(sizes);

    /* Give the window a nice icon */
    QPixmap pixmap(icon_64x64_xpm);
    QIcon icon(pixmap);

    this->setWindowIcon(icon);

#ifdef Q_OS_MAC
    /* Register the file association for OSX (other platforms use a different method) */
    connect(app, SIGNAL(loadFile(QString)), SLOT(loadNZBFile(QString)));
#endif

    connect(ui->actionImportNZBFile, SIGNAL(triggered(void)),
            SLOT(actionImportNZB_activated(void)));
    connect(ui->actionPauseDownloadQueue, SIGNAL(triggered(void)),
            SLOT(actionPauseDownloadQueue_activated()));
    connect(ui->actionDonate, SIGNAL(triggered(void)),
            SLOT(actionDonate_activated(void)));
    connect(ui->actionEnableAutomaticShutdown, SIGNAL(triggered(void)),
            SLOT(actionEnableAutomaticShutdown_activated()));

    connect(ui->actionManagePlugins, SIGNAL(triggered(void)),
            SLOT(actionManagePlugins_activated(void)));
    connect(ui->actionShowDebugWindow, SIGNAL(triggered(void)),
            SLOT(actionShowDebugWindow_activated(void)));

    connect(glue, SIGNAL(fileStateChanged(QString,QString,QString,QNNTPGrabGlue::TaskState,QNNTPGrabGlue::TaskState)),
            SLOT(onFileStateChanged(QString,QString,QString,QNNTPGrabGlue::TaskState,QNNTPGrabGlue::TaskState)));
    connect(glue, SIGNAL(schedularStateChanged(QNNTPGrabGlue::SchedularState,QString)),
            SLOT(onSchedularStateChanged(QNNTPGrabGlue::SchedularState,QString)));
    if (!glue->getIsStandalone()) {
        connect(glue, SIGNAL(connectionLost(QString)), SLOT(onConnectionLost(QString)));
    }

    connect(app, SIGNAL(messageReceived(QString)), SLOT(onMessageReceived(QString)));
    app->setActivationWindow(this, true);

    if (glue->getIsStandalone()) {
        glue->startSchedular();
    } else {
        onSchedularStateChanged(glue->getSchedularState(), "");
    }

    /* Load all plugins which should be loaded automatically */
    ConfigGuiOpts opts;
    QStringList auto_load_plugins = opts.getAutoLoadPlugins();
    Q_FOREACH(QString plugin_name, auto_load_plugins) {
        QNNTPGrabGlue::PluginInfo plugin_info;

        if (!glue->getPluginInfo(plugin_name, plugin_info) || plugin_info.is_loaded) {
            /* Plugin not found or already loaded, ignore */
            continue;
        }

        QString errmsg;
        if (!glue->loadPlugin(plugin_name, errmsg)) {
            qWarning((tr("Unable to load plugin '%1': %2").arg(plugin_name).arg(errmsg)).toUtf8());
        }
    }

    /* Create a tray icon if possible */
    if (QSystemTrayIcon::isSystemTrayAvailable()) {
        trayIcon.setIcon(icon);
        trayIcon.setToolTip("NNTPGrab");
        connect(&trayIcon,
                SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
                SLOT(trayIconActivated(QSystemTrayIcon::ActivationReason)));

        /* Show the tray icon if requested */
        if (opts.getShowTrayIcon()) {
            showTrayIcon();
        }
    }

    bool beenhere = false;
    Q_FOREACH(QString arg, QApplication::arguments()) {
        /* Ignore the first argument (which is the executable) */
        if (!beenhere) {
            beenhere = true;
            continue;
        }

        /* Test if the argument is a file name */
        QFile f(arg);
        if (f.exists()) {
            /* Try to open it */
            loadNZBFile(arg);
        }
    }
}

MainWindow::~MainWindow()
{
    delete updateInfoWidget;
    delete configWidget;
    delete downloadQueueWidget;
    delete nzbcreatorWidget;
    delete repairAndUnpackWidget;

    delete trafficMonitorWidget;
    delete connectionsWidget;

    delete ui;

    if (formDebug) {
        delete formDebug;
    }

    if (glue) {
        delete glue;
    }
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::actionImportNZB_activated(void)
{
    DialogImportNZB d(glue);
    d.exec();
}

void MainWindow::actionDonate_activated(void)
{
    QDesktopServices::openUrl(QUrl("http://www.nntpgrab.nl/donate.php"));
}

void MainWindow::actionEnableAutomaticShutdown_activated(void)
{
    if (ui->actionEnableAutomaticShutdown->isChecked()) {
        connect(glue, SIGNAL(allDownloadsCompleted()), SLOT(onAllDownloadsCompleted()));
    } else {
        disconnect(this, SLOT(onAllDownloadsCompleted()));
    }
}

void MainWindow::actionManagePlugins_activated(void)
{
    DialogManagePlugins d(glue);
    d.exec();
}

void MainWindow::actionShowDebugWindow_activated(void)
{
    if (formDebug) {
        formDebug->show();
        formDebug->activateWindow();
        return;
    }

    formDebug = new FormDebug(glue);
    formDebug->show();
}

void MainWindow::actionPauseDownloadQueue_activated(void)
{
    if (!ui->actionPauseDownloadQueue->isChecked()) {
        /* The user just pressed the button to indicate that the schedular needs to be stopped */
        if (glue->getSchedularState() == QNNTPGrabGlue::SCHEDULAR_STATE_STOPPED) {
            glue->startSchedular();
        } else {
            qWarning(tr("The schedular is currently being stopped").toUtf8().data());
        }
    } else {
        glue->stopSchedular(false);
        ui->actionPauseDownloadQueue->setEnabled(false);
    }
}

void MainWindow::onFileStateChanged(QString collection_name, QString subject, QString real_filename, QNNTPGrabGlue::TaskState old_state, QNNTPGrabGlue::TaskState new_state)
{
    Q_UNUSED(collection_name);
    Q_UNUSED(real_filename);
    Q_UNUSED(old_state);

    QString msg;
    QString title;

    switch (new_state) {
        case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DECODE:
            if (!showNotificationOnFileDownloaded) {
                return;
            }

            msg = tr("All the parts belonging to the post with subject\n%1\nare now downloaded and will soon be decoded").arg(subject);
            title = tr("File download done");
            break;

        case QNNTPGrabGlue::TASK_STATE_FINISHED_COMPLETE:
        case QNNTPGrabGlue::TASK_STATE_FINISHED_INCOMPLETE:
        case QNNTPGrabGlue::TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            if (!showNotificationOnFileDecoded) {
                return;
            }

            msg = tr("The file belonging to the post with subject\n%1\nis now decoded").arg(subject);
            title = tr("File decode done");
            break;

        case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DOWNLOAD:
        case QNNTPGrabGlue::TASK_STATE_DOWNLOADING:
        case QNNTPGrabGlue::TASK_STATE_DECODING:
        case QNNTPGrabGlue::TASK_STATE_SKIPPED:
        default:
            return;
            break;
    };

    showNotification(title, msg);
}

void MainWindow::onSchedularStateChanged(QNNTPGrabGlue::SchedularState state, QString reason)
{
    if (state == QNNTPGrabGlue::SCHEDULAR_STATE_RUNNING) {
        ui->actionPauseDownloadQueue->setChecked(false);
        ui->actionPauseDownloadQueue->setEnabled(true);
        ui->actionPauseDownloadQueue->setText(tr("Pause download queue"));
    } else if (state == QNNTPGrabGlue::SCHEDULAR_STATE_STOPPED) {
        ui->actionPauseDownloadQueue->setChecked(true);
        ui->actionPauseDownloadQueue->setEnabled(true);
        ui->actionPauseDownloadQueue->setText(tr("Resume download queue"));
    }

    if (state == QNNTPGrabGlue::SCHEDULAR_STATE_STOPPING && !reason.isEmpty()) {
        qWarning(tr("The schedular has been paused due to an error:\n%s").toUtf8().data(), reason.toUtf8().data());
        ui->actionPauseDownloadQueue->setEnabled(false);
    }
}

void MainWindow::onConnectionLost(QString errmsg)
{
    qWarning(tr("The connection with the NNTPGrab Server was lost: %1").arg(errmsg).toUtf8());
    close();
}

void MainWindow::showNotification(QString title, QString msg)
{
#ifdef Q_OS_MAC
    /* Due to QTBUG-7638 a tray icon needs to be shown on
     * OSX before a Growl notification can be shown */
    bool wasVisible = trayIcon.isVisible();
    if (!wasVisible) {
        trayIcon.show();
    }
#endif

    trayIcon.showMessage(title, msg);

#ifdef Q_OS_MAC
    if (!wasVisible) {
        trayIcon.hide();
    }
#endif
}

void MainWindow::setShowNotificationOnFileDownloaded(bool enabled)
{
    showNotificationOnFileDownloaded = enabled;
}

void MainWindow::setShowNotificationOnFileDecoded(bool enabled)
{
    showNotificationOnFileDecoded = enabled;
}

void MainWindow::onMessageReceived(QString msg)
{
    if (msg == "wakeUp") {
        this->show();
        this->raise();
        this->activateWindow();
        return;
    }

    /* Check if the message is a file name */
    QFile f(msg);
    if (f.exists()) {
        DialogImportNZB d(glue);
        d.readNZBFile(msg);
        d.exec();
    }
}

void MainWindow::onAllDownloadsCompleted(void)
{
    qDebug("All downloads are completed. The program will now shut down automatically\n");
    if (!glue->utilsPerformShutdown()) {
        qWarning("Unknown error occured while trying to shut down the computer");
    } else {
        close();
    }
}

void MainWindow::loadNZBFile(QString filename)
{
    DialogImportNZB d(glue);
    d.readNZBFile(filename);
    d.exec();
}

void MainWindow::showTrayIcon()
{
    QPixmap pixmap(icon_64x64_xpm);
    QIcon icon(pixmap);

    trayIcon.show();
    trayIconIsShown = true;
}

void MainWindow::hideTrayIcon()
{
    trayIcon.hide();
    trayIconIsShown = false;
}

void MainWindow::trayIconActivated(QSystemTrayIcon::ActivationReason reason)
{
    if (reason == QSystemTrayIcon::Trigger) {
        if (this->isVisible() && this->isActiveWindow()) {
            /* If we're already being shown AND on the foreground, hide */
            this->hide();
        } else if (this->isVisible() && !this->isActiveWindow()) {
            /* If we're already being shown AND NOT on the foreground, bring to the foreground */
            this->raise();
            this->activateWindow();
        } else if (!this->isVisible()) {
            /* If we're NOT being shown, show */
            this->show();
            this->raise();
            this->activateWindow();
        }
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    Q_UNUSED(event);

    /* Automatically close any other dialogs if they're open */
    if (formDebug) {
        delete formDebug;
        formDebug = NULL;
    }

    QMainWindow::closeEvent(event);
}

#ifdef Q_WS_WIN
#include <windows.h>
bool MainWindow::winEvent(MSG* msg, long* result)
{
    if (!trayIconIsShown) {
        return false;
    }

    if (msg->message == WM_SYSCOMMAND && msg->wParam == SC_MINIMIZE) {
        hide(); // minimized toolwindows look bad on Windows, so let's just hide it instead
                // plus we cannot do this in changeEvent(), because it's called too late
        *result = 0;
        return true;    // don't let Qt process this event
    }
    return false;
}
#elif defined(Q_WS_X11)
void MainWindow::hideEvent(QHideEvent *e)
{
    QMainWindow::hideEvent(e);

    if (!trayIconIsShown) {
        return;
    }

    /* Hide the program from the task bar when the tray icon is visible */
    if (QSystemTrayIcon::isSystemTrayAvailable()) {
        this->hide();
    }
}
#endif
