/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <glib.h>
#ifdef WIN32
#include <winsock2.h>
#endif
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_plugin.h"
#include "download_queue.h"
#include "nntpconnection.h"
#include "configuration.h"
#include "decoder_thread.h"
#include "download_thread.h"

#if !GLIB_CHECK_VERSION(2,20,0) && !defined(WIN32)
/* The g_poll function was added in GLib 2.20.0.
 * Provide an implementation of this function for environments
 * which have an older version of GLib installed.
 * Source: http://git.gnome.org/browse/glib/tree/glib/gpoll.c
 */
#include <poll.h>

static gint
g_poll (GPollFD *fds, guint nfds, gint timeout)
{
    return poll ((struct pollfd *)fds, nfds, timeout);
}
#endif

#ifdef WIN32
/* The g_poll function only works on HANDLE's on Win32
 * so we replace the GLib implementation with one that
 * uses the select() function instead of poll().
 * The code below is taken from glib's gpoll.c where
 * it is a fallback path for platforms which don't
 * have a poll function() */
#undef g_poll
#define g_poll win32_g_poll
static gint
win32_g_poll (GPollFD *fds, guint nfds, gint timeout)
{
    struct timeval tv;
    FD_SET rset, wset, xset;
    GPollFD *f;
    int ready;
    int maxfd = 0;

    FD_ZERO (&rset);
    FD_ZERO (&wset);
    FD_ZERO (&xset);

    for (f = fds; f < &fds[nfds]; ++f)
        if (f->fd >= 0)
        {
            if (f->events & G_IO_IN)
                FD_SET (f->fd, &rset);
            if (f->events & G_IO_OUT)
                FD_SET (f->fd, &wset);
            if (f->events & G_IO_PRI)
                FD_SET (f->fd, &xset);
            if (f->fd > maxfd && (f->events & (G_IO_IN|G_IO_OUT|G_IO_PRI)))
                maxfd = f->fd;
        }

    tv.tv_sec = timeout / 1000;
    tv.tv_usec = (timeout % 1000) * 1000;

    ready = select (maxfd + 1, &rset, &wset, &xset,
                    timeout == -1 ? NULL : &tv);
    if (ready > 0)
        for (f = fds; f < &fds[nfds]; ++f)
        {
            f->revents = 0;
            if (f->fd >= 0)
            {
                if (FD_ISSET (f->fd, &rset))
                    f->revents |= G_IO_IN;
                if (FD_ISSET (f->fd, &wset))
                    f->revents |= G_IO_OUT;
                if (FD_ISSET (f->fd, &xset))
                    f->revents |= G_IO_PRI;
            }
        }

    return ready;
}

static void
strip_newline(char *line)
{
    if (line[strlen(line) - 1]  == '\n') {
        line[strlen(line) - 1] = '\0';
    }

    if (line[strlen(line) - 1]  == '\r') {
        line[strlen(line) - 1] = '\0';
    }
}

static char *
WSAGetStrError(int err)
{
    LPVOID lpMsgBuf;
    char *ret;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        (DWORD) err,
        0,
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    ret = g_strdup(lpMsgBuf);
    strip_newline(ret);

    LocalFree(lpMsgBuf);

    return ret;
}

#else
#define O_BINARY 0
#endif

typedef enum {
    INTERNAL_SCHEDULAR_STATE_RUNNING,
    INTERNAL_SCHEDULAR_STATE_STOPPING,
    INTERNAL_SCHEDULAR_STATE_STOPPED,
    INTERNAL_SCHEDULAR_STATE_SUSPENDING,
    INTERNAL_SCHEDULAR_STATE_SUSPENDED
} InternalSchedularState;

static GStaticMutex thread_mutex = G_STATIC_MUTEX_INIT;
static GThread *download_thread = NULL;
static InternalSchedularState download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPED;
static char abort_reason[1024];
static Configuration *config = NULL;
static ngboolean decoder_is_active = FALSE;

static NNTPConnectionInfo *
lookup_connection(GList *active_connections, GPollFD poll_fd)
{
    GList *list = active_connections;

    while (list) {
        NNTPConnectionInfo *conn = list->data;

        if (conn->poll_fd.fd == poll_fd.fd) {
            return conn;
        }

        list = g_list_next(list);
    }

    return NULL;
}

static void
reset_poll_fd_list(GList *active_connections, GPollFD **active_connection_polls, int *num_active_connections)
{
    GList *list;
    GPollFD *new_polls = NULL;
    int count;

    g_return_if_fail(active_connections != NULL);
    g_return_if_fail(active_connection_polls != NULL);
    g_return_if_fail(num_active_connections != NULL);

    /* First calculate the number of really active connections */
    list = active_connections;
    count = 0;
    while (list) {
        NNTPConnectionInfo *conn = list->data;

        if (conn->poll_fd.fd >= 0) {
            count++;
        }

        list = g_list_next(list);
    }

    /* Free the old list of GPollFD's */
    g_slice_free1(sizeof(GPollFD) * *num_active_connections, *active_connection_polls);

    if (count == 0) {
        /* No active connections. Update the pointers */
        *active_connection_polls = NULL;
        *num_active_connections = 0;
        return;
    }

    /* Build a new list of GPollFD's */
    new_polls = g_slice_alloc0(sizeof(GPollFD) * count);

    list = active_connections;
    count = 0;
    while (list) {
        NNTPConnectionInfo *conn = list->data;

        if (conn->poll_fd.fd >= 0) {
            new_polls[count] = conn->poll_fd;
            count++;
        }

        list = g_list_next(list);
    }

    /* Update pointers */
    *active_connection_polls = new_polls;
    *num_active_connections = count;
}

static void
drop_socket(GList **active_connections, GPollFD **active_connection_polls, int *num_active_connections, GPollFD poll_fd, NNTPConnectionInfo *conn)
{
    /* Kill any file descriptor if it's still opened */
    if (conn->article_fd >= 0) {
        close(conn->article_fd);
    }

    reset_poll_fd_list(*active_connections, active_connection_polls, num_active_connections);

    /* Reset the item */
    conn->poll_fd.fd = -1;
    conn->is_idle = TRUE;
    conn->idle_start_stamp = 0;
}

static void
disconnect_socket(GList **active_connections, GPollFD **active_connection_polls, int *num_active_connections, GPollFD poll_fd, NNTPDisconnectType disconnect_type, const char *reason, int cause_line)
{
    NNTPConnectionInfo *conn;

    g_return_if_fail(active_connections != NULL);
    g_return_if_fail(*active_connections != NULL);
    g_return_if_fail(active_connection_polls != NULL);
    g_return_if_fail(*active_connection_polls != NULL);
    g_return_if_fail(num_active_connections != NULL);
    g_return_if_fail(*num_active_connections > 0);
    g_return_if_fail(poll_fd.fd >= 0);

    /* Shut down the network connection */
    conn = lookup_connection(*active_connections, poll_fd);
    g_return_if_fail(conn != NULL);
    nntpconnection_disconnect_from_server(conn, disconnect_type, reason, __FILE__, cause_line);

    drop_socket(active_connections, active_connection_polls, num_active_connections, poll_fd, conn);
}

static gpointer
download_thread_func(gpointer data)
{
    int i;
    int retval;
    GList *active_connections = NULL;  /* NNTPConnectionInfo */
    int num_active_connections = 0;
    GPollFD *active_connection_polls = NULL;
    GList *avail_servers, *list;
    NGConfigOpts opts;
    time_t curr_time;
    int server_id;
    int prev_server_id;
    int connection_id;
    time_t stamp_last_queued_item_found = time(NULL);
    gboolean nothing_to_download;
    gboolean connection_error_occured = FALSE;

    g_return_val_if_fail(config != NULL, NULL);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "download_thread_func: initializing");

    memset(abort_reason, 0, sizeof(abort_reason));

    /* Initialize the decoder if it wasn't started already */
    if (!decoder_is_active) {
        decoder_thread_start();
        decoder_is_active = TRUE;
    }

    /* Build a list of allowed connections */
    connection_id = 1;
    server_id = 0;
    avail_servers = configuration_get_avail_servers(config);
    opts = configuration_get_opts(config);
    list = avail_servers;
    while (list) {
        const char *servername = list->data;
        NGConfigServer *server = configuration_get_server_info(config, servername);

        /* Ignore disabled servers */
        if (!server->enabled) {
            server_id++;
            list = g_list_next(list);
            g_slice_free(NGConfigServer, server);
            continue;
        }

        for (i = 1; i <= server->max_threads; i++) {
            NNTPConnectionInfo *conn = g_slice_new0(NNTPConnectionInfo);

            conn->connection_id = connection_id;
            connection_id++;
            conn->server_id = server_id;
            conn->poll_fd.fd = -1;
            conn->is_idle = TRUE;
            conn->article_fd = -1;
            conn->recv_buffer_length = 0;
            memcpy(&conn->server_info, server, sizeof(NGConfigServer));
            if (opts.enable_bandwidth_shaping) {
                conn->max_bandwidth = opts.max_bandwidth;
            }

            active_connections = g_list_append(active_connections, conn);
        }

        g_slice_free(NGConfigServer, server);

        server_id++;
        list = g_list_next(list);
    }
    configuration_free_avail_servers(config, avail_servers);
    avail_servers = NULL;

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "download_thread_func: Prepared a list for maximum %i connections", g_list_length(active_connections));

    /* Download thread main loop */
    while (TRUE) {
        /* Is the schedular still running? */
        g_static_mutex_lock(&thread_mutex);
        if (download_thread_state != INTERNAL_SCHEDULAR_STATE_RUNNING) {
            g_static_mutex_unlock(&thread_mutex);
            break;
        }
        g_static_mutex_unlock(&thread_mutex);

        /* When there are idle/inactive connections, check the download queue whether we need to download something */
        curr_time = time(NULL);
        prev_server_id = -1;
        list = active_connections;
        while (list) {
            NNTPConnectionInfo *conn = list->data;

            /* Sanity check */
            g_return_val_if_fail(conn, NULL);

            /* Is this connection already doing something ? */
            if (!conn->is_idle) {
                /* Guard against read timeouts */
                if (conn->last_activity_stamp + 30 < curr_time) {
                    /* No activity happened in the last 30 seconds, disconnect from the server */
                    disconnect_socket(&active_connections, &active_connection_polls, &num_active_connections, conn->poll_fd, DISCONNECT_READ_TIMEOUT, NULL, __LINE__);

                    /* This socket is ready for re-use, so try again directly */
                    continue;
                }

                list = g_list_next(list);
                continue;
            }

            /* Was this socket recently disconnected? (Don't hammer the server) */
            if (conn->last_disconnect_stamp + 30 > curr_time) {
                /* Try again later */
                connection_error_occured = TRUE;
                list = g_list_next(list);
                continue;
            } else {
                connection_error_occured = FALSE;
            }

            /* Have we already checked this connection recently? */
            if (conn->last_checked_stamp == curr_time) {
                /* Yes, ignore for now */
                list = g_list_next(list);
                continue;
            }

            /* Only check servers where were not sure if new tasks are ready */
            if (prev_server_id == conn->server_id) {
                /* Automatically disconnect if we've been idle for too long */
                if (conn->poll_fd.fd >= 0 && conn->idle_start_stamp + 30 < curr_time) {
                    /* Sanity check */
                    g_return_val_if_fail(conn->poll_fd.fd >= 0, NULL);

                    /* Kill the connection */
                    disconnect_socket(&active_connections, &active_connection_polls, &num_active_connections, conn->poll_fd, DISCONNECT_NORMAL, NULL, __LINE__);
                }

                list = g_list_next(list);
                continue;
            }

            /* Do we have anything to download from this server? */
            if (!download_queue_get_next_part_to_download(&conn->collection, &conn->file, &conn->part, conn->server_id, &nothing_to_download)) {
                /* Is the download queue empty or is everything downloaded? */
                if (nothing_to_download) {
                    /* Go to sleep */
                    g_static_mutex_lock(&thread_mutex);
                    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "All downloads are completed or there's nothing to download. Now entering suspend mode");
                    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_RUNNING) {
                        download_thread_state = INTERNAL_SCHEDULAR_STATE_SUSPENDING;
                    }
                    g_static_mutex_unlock(&thread_mutex);
                    break;
                }

                /* Ignore this server for the next iteration */
                prev_server_id = conn->server_id;

                /* Nothing to download. Update our timer */
                conn->last_checked_stamp = curr_time;

                /* Automatically disconnect when we've been idle for more than 30 seconds */
                if (conn->poll_fd.fd >= 0 && conn->idle_start_stamp + 30 < curr_time) {
                    /* Sanity check */
                    g_return_val_if_fail(conn->poll_fd.fd >= 0, NULL);

                    /* Kill the connection */
                    disconnect_socket(&active_connections, &active_connection_polls, &num_active_connections, conn->poll_fd, DISCONNECT_NORMAL, NULL, __LINE__);
                }
            } else {
                /* Get ready to download something */
                stamp_last_queued_item_found = time(NULL);
                conn->job_type = NNTP_JOB_TYPE_ARTICLE;

                /* Generate the filename where we can download data to */
                g_return_val_if_fail(strlen(conn->file->tmp_filename) > 0, NULL);

                memset(conn->article_filename, 0, sizeof(conn->article_filename));
                snprintf(conn->article_filename, sizeof(conn->article_filename) - 1, "%s%s%s.%i", opts.temp_directory, G_DIR_SEPARATOR_S, conn->file->tmp_filename, conn->part->part_num);

                /* We may need to check for new tasks later during this run, so reset the prev_server_id here */
                prev_server_id = -1;

                /* Create a connection if necessary */
                if (conn->poll_fd.fd > -1) {
                    nntpconnection_send_body_command(conn);
                    conn->is_idle = FALSE;
                } else {
                    char *errmsg = NULL;
                    NNTPConnectionErrCode retval = nntpconnection_connect_to_server(conn, &errmsg);

                    if (retval != NNTP_CONNECTION_ERROR_NONE) {
                        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_connect_to_server FAILED: %s (%i)", errmsg, retval);
                        if (conn->collection && conn->file && conn->part) {
                            /* If a connection timeout occurs the status will already be adjusted */
                            download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);
                        }

                        /* If the hostname could not be resolved, bail out as this indicates a configuration error */
                        if (retval == NNTP_CONNECTION_ERROR_HOST_NOT_FOUND) {
                            g_static_mutex_lock(&thread_mutex);

                            snprintf(abort_reason, sizeof(abort_reason) - 1, _("Unable to detect the IP address belonging to '%s' (servername = %s)"), conn->server_info.hostname, conn->server_info.servername);
                            download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPING;
                            nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPING, abort_reason);

                            g_static_mutex_unlock(&thread_mutex);

                            break;
                        } else {
                            connection_error_occured = TRUE;
                        }
                    } else {
                        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_connect_to_server SUCCEEDED, conn_id = %i, poll_fd = %i", conn->connection_id, conn->poll_fd.fd);
                        reset_poll_fd_list(active_connections, &active_connection_polls, &num_active_connections);
                    }
                }
            }

            list = g_list_next(list);
        }

        /* Are there any active connections and no retry attempts ? */
        if (num_active_connections == 0) {
            if (connection_error_occured) {
                /* If the server disconnected by any reason, try again later */
                g_usleep(G_USEC_PER_SEC * 1);
                continue;
            } else {
                /* Nothing to do, abort this thread and enter suspend mode */
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Download thread has nothing to do. Now entering suspend mode");
                g_static_mutex_lock(&thread_mutex);
                if (download_thread_state == INTERNAL_SCHEDULAR_STATE_RUNNING) {
                    download_thread_state = INTERNAL_SCHEDULAR_STATE_SUSPENDING;
                }
                g_static_mutex_unlock(&thread_mutex);
                break;
            }
        }

        /* Check for new data on the active connections */
        g_return_val_if_fail(active_connection_polls != NULL, NULL);

        /* Reset old poll results */
        for (i = 0; i < num_active_connections; i++) {
            active_connection_polls[i].events = G_IO_IN | G_IO_HUP | G_IO_ERR;
            active_connection_polls[i].revents = 0;
        }

        /* Perform the poll for 1 second (1000msec) */
        retval = g_poll(active_connection_polls, num_active_connections, 1000);

        switch (retval) {
            case -1:    /* Call was interrupted, see errno for details */
                if (errno == EINTR) {
                    continue;   /* Try again */
                } else {
#ifdef WIN32
                    char *errmsg = WSAGetStrError(WSAGetLastError());
                    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, __FILE__ ":%i Unknown error occured while polling: %s\n", __LINE__, errmsg);
                    g_free(errmsg);
#else
                    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, __FILE__ ":%i Unknown error occured while polling: %s\n", __LINE__, strerror(errno));
#endif
                    return FALSE;
                }

            case 0:     /* Timeout */
                continue;

            default:    /* Some activity occured */
                for (i = 0; i < num_active_connections; i++) {
                    g_return_val_if_fail(i >= 0, NULL);

                    GPollFD *poll_fd = &active_connection_polls[i];
                    if (poll_fd->revents == G_IO_ERR || poll_fd->revents == G_IO_HUP || poll_fd->revents == G_IO_NVAL) {
                        disconnect_socket(&active_connections, &active_connection_polls, &num_active_connections, *poll_fd, DISCONNECT_READ_ERROR, _("Unexpected error occured while polling connection"), __LINE__);
                        i--;    /* As one element has been removed we need to test the current position again */
                    } else if (poll_fd->revents == G_IO_IN) {
                        NNTPConnectionInfo *conn = lookup_connection(active_connections, *poll_fd);

                        g_return_val_if_fail(conn != NULL, NULL);

                        nntpconnection_process_socket_activity(conn);

                        /* A disconnect could have occured when processing the socket activity */
                        if (conn->poll_fd.fd == -1) {
                            drop_socket(&active_connections, &active_connection_polls, &num_active_connections, *poll_fd, conn);
                        }
                    }
                }
        }
    }

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "download_thread_func: Request received to abort thread");

    /* The user has emit a request to abort the download thread. Perform a nice cleanup */
    /* NOTE: this cleanup needs to happen in reverse order! */
    for (i = num_active_connections; i > 0; i--) {
        g_return_val_if_fail(active_connection_polls != NULL, NULL);
        g_return_val_if_fail(num_active_connections > 0, NULL);
        disconnect_socket(&active_connections, &active_connection_polls, &num_active_connections, active_connection_polls[i - 1], DISCONNECT_NORMAL, NULL, __LINE__);
    }

    list = active_connections;
    while (list) {
        NNTPConnectionInfo *conn = list->data;

        g_slice_free(NNTPConnectionInfo, conn);

        list = g_list_next(list);
    }

    /* Don't stop the decoder when the download queue got auto-suspended
       (all files are downloaded, but some files may still need to be decoded) */
    g_static_mutex_lock(&thread_mutex);
    if (download_thread_state != INTERNAL_SCHEDULAR_STATE_SUSPENDING) {
        decoder_thread_stop();
        decoder_is_active = FALSE;
    }

    /* If the download thread was stopped by outside, emit a message to the frontend */
    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_STOPPING) {
        download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPED;
        nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPED, abort_reason);
    }

    download_thread = NULL;
    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDING) {
        download_thread_state = INTERNAL_SCHEDULAR_STATE_SUSPENDED;
    }

    g_static_mutex_unlock(&thread_mutex);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "download_thread_func: Download thread aborted and cleaned up");

    return NULL;
}

static gboolean
on_config_changed(Configuration *config, gpointer data)
{
    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_RUNNING ||
        download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDING ||
        download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDED) {

        /* Restart the download thread to use new settings */
        download_thread_stop(TRUE, NULL);
        download_thread_start(config);
    }

    return FALSE;
}

static gboolean
download_thread_start_real(void)
{
    g_return_val_if_fail(download_thread == NULL, FALSE);

    g_static_mutex_lock(&thread_mutex);

    memset(abort_reason, 0, sizeof(abort_reason));

    download_thread_state = INTERNAL_SCHEDULAR_STATE_RUNNING;
    download_thread = g_thread_create(download_thread_func, NULL, TRUE, NULL);

    if (!download_thread) {
        download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPED;
        g_static_mutex_unlock(&thread_mutex);

        return FALSE;
    }

    g_static_mutex_unlock(&thread_mutex);

    return TRUE;
}

gboolean
download_thread_start(Configuration *configuration)
{
    static gboolean beenhere = FALSE;

    g_return_val_if_fail(download_thread == NULL, FALSE);
    g_return_val_if_fail(download_thread_state == INTERNAL_SCHEDULAR_STATE_STOPPED, FALSE);
    g_return_val_if_fail(configuration != NULL, FALSE);

    config = configuration;

    if (!beenhere) {
        beenhere = TRUE;
        g_signal_connect(config, "config_changed", G_CALLBACK(on_config_changed), NULL);
    }

    if (!download_thread_start_real()) {
        g_return_val_if_fail(download_thread_state == INTERNAL_SCHEDULAR_STATE_STOPPED, FALSE);
        return FALSE;
    }

    nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_RUNNING, NULL);

    return TRUE;
}

gboolean
download_thread_stop(gboolean dont_stop_decoder, const char *reason)
{
    while (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDING) {
        /* Wait for a bit until the suspend has completed */
        g_usleep(G_USEC_PER_SEC / 10);
    }

    g_static_mutex_lock(&thread_mutex);

    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDED) {
        /* Download thread is already stopped because there are no new tasks */
        g_return_val_if_fail(download_thread == NULL, FALSE);

        nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPING, reason);
        download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPED;
        nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPED, reason);

        /* Clean up the decoder if it's still running */
        if (decoder_is_active && !dont_stop_decoder) {
            decoder_thread_stop();
            decoder_is_active = FALSE;
        }

        g_static_mutex_unlock(&thread_mutex);

        return TRUE;
    }

    /* Perform a normal cleanup */
    g_return_val_if_fail(download_thread != NULL, FALSE);

    if (!reason) {
        reason = "";
    }

    strncpy(abort_reason, reason, sizeof(abort_reason) - 1);
    download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPING;
    nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPING, reason);

    g_static_mutex_unlock(&thread_mutex);

    /* Wait for the download thread to perform it's cleanup */
    g_thread_join(download_thread);

    /* download_thread_state should be set to INTERNAL_SCHEDULAR_STATE_STOPPED by now */
    g_return_val_if_fail(download_thread_state == INTERNAL_SCHEDULAR_STATE_STOPPED, FALSE);

    return TRUE;
}

NGSchedularState
download_thread_get_state (void)
{
    switch (download_thread_state) {
        case INTERNAL_SCHEDULAR_STATE_RUNNING:
        case INTERNAL_SCHEDULAR_STATE_SUSPENDING:
        case INTERNAL_SCHEDULAR_STATE_SUSPENDED:
            return SCHEDULAR_STATE_RUNNING;

        case INTERNAL_SCHEDULAR_STATE_STOPPING:
            return SCHEDULAR_STATE_STOPPING;

        case INTERNAL_SCHEDULAR_STATE_STOPPED:
        default:
             return SCHEDULAR_STATE_STOPPED;
    }
}

void
download_thread_abort_without_waiting(const char *format, ...)
{
    va_list ap;

    while (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDING) {
        /* Wait for a bit until the suspend has completed */
        g_usleep(G_USEC_PER_SEC / 10);
    }

    g_static_mutex_lock(&thread_mutex);

    if (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDED) {
        /* Download thread is already stopped because there are no new tasks */
        g_return_if_fail(download_thread == NULL);

        nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPING, NULL);
        download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPED;
        nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPED, NULL);

        /* Clean up the decoder if it's still running */
        if (decoder_is_active) {
            decoder_thread_stop();
            decoder_is_active = FALSE;
        }

        g_static_mutex_unlock(&thread_mutex);

        return;
    }

    g_return_if_fail(format != NULL);

    va_start(ap, format);
    vsnprintf(abort_reason, sizeof(abort_reason) - 1, format, ap);
    va_end(ap);

    download_thread_state = INTERNAL_SCHEDULAR_STATE_STOPPING;
    nntpgrab_core_emit_schedular_state_changed(FALSE, SCHEDULAR_STATE_STOPPING, abort_reason);

    g_static_mutex_unlock(&thread_mutex);
}

void
download_thread_awaken_if_needed(void)
{
    while (download_thread_state == INTERNAL_SCHEDULAR_STATE_SUSPENDING) {
        /* Wait for a bit until the suspend has completed */
        g_usleep(G_USEC_PER_SEC / 10);
    }

     if (download_thread_state != INTERNAL_SCHEDULAR_STATE_SUSPENDED) {
         return;
     }

     if (!download_thread_start_real()) {
         g_return_if_fail(download_thread_state == INTERNAL_SCHEDULAR_STATE_STOPPED);
    } else {
         g_return_if_fail(download_thread_state == INTERNAL_SCHEDULAR_STATE_RUNNING);
    }
}

