/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"
#include "marshalers.h"
#include "configuration.h"
#include "nntpconnection.h"
#include "collection_alloc.h"
#include "download_thread.h"
#include "download_queue.h"
#include "decoder_thread.h"
#include "plugins.h"

#ifdef WIN32
#define tempnam _tempnam
#endif

struct NntpgrabCore
{
    GObject parent;

    gboolean is_initialized;
    Configuration *config;
    gboolean is_finalizing;
};

static NntpgrabCore *core_global = NULL;

struct NntpgrabCoreClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    PART_DOWNLOAD_START_SIGNAL,
    PART_DONE_SIGNAL,
    PART_FAILED_SIGNAL,
    PART_PROGRESS_UPDATE_SIGNAL,
    TRAFFIC_MONITOR_UPDATE_SIGNAL,
    COLLECTION_ADDED_SIGNAL,
    COLLECTION_REMOVED_SIGNAL,
    COLLECTION_MODIFIED_SIGNAL,
    FILE_ADDED_SIGNAL,
    FILE_REMOVED_SIGNAL,
    FILE_DOWNLOAD_STATE_UPDATE_SIGNAL,
    FILE_STATE_CHANGED_SIGNAL,
    CONNECTION_CONNECTING_SIGNAL,
    CONNECTION_CONNECTED_SIGNAL,
    CONNECTION_DISCONNECT_SIGNAL,
    SCHEDULAR_STATE_CHANGED_SIGNAL,
    LOG_MESSAGE_SIGNAL,
    TASK_MOVED_SIGNAL,
    COLLECTION_MOVED_SIGNAL,
    COLLECTION_DOWNLOADED_SIGNAL,
    ALL_DOWNLOADS_COMPLETED_SIGNAL,
    PLUGIN_LOADED_SIGNAL,
    PLUGIN_UNLOADED_SIGNAL,
    PLUGIN_EVENT_SIGNAL,
    QUIT_REQUESTED_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(NntpgrabCore, core, G_TYPE_OBJECT)

NGType
nntpgrab_core_get_type(void)
{
    return (NGType) core_get_type();
}

static void
core_init (NntpgrabCore *obj)
{
    obj->is_finalizing = FALSE;
    obj->is_initialized = FALSE;
}

static void
core_finalize (GObject *obj)
{
    char *errmsg = NULL;
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    core->is_finalizing = TRUE;

    if (!core->is_initialized) {
        return;
    }

    if (download_thread_get_state() == SCHEDULAR_STATE_RUNNING) {
        download_thread_stop(FALSE, NULL);
    }

    if (!download_queue_destroy(&errmsg)) {
        g_print(__FILE__ ":%i Unable to clean up the download queue: %s\n", __LINE__, errmsg);
        g_free(errmsg);
    }

    decoder_thread_destroy();
    plugins_destroy();

    core_signals_clear_stack();

    if (core->config) {
        configuration_destroy(core->config);
    }

    core_global = NULL;
}

static void
core_class_init (NntpgrabCoreClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = core_finalize;

#define create_signal(a,b,c,d,e,f,g,h,i, j...)                    \
            g_signal_new(a,b,c,d,e,f,g,h,i, ## j);                \
            ng_plugin_create_internal_event(a, g, h, i, ## j);

    plugins_initialize();

    /**
     * NntpgrabCore::config-changed:
     * @obj: An instance of the NntpgrabCore
     *
     * Something in the configuration has changed
     */
    signals[CONFIG_CHANGED_SIGNAL] =      create_signal("config_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__VOID,
                                                        G_TYPE_NONE, 0);

    /**
     * NntpgrabCore::part-download-start:
     * @obj:              An instance of the NntpgrabCore
     * @servername:       The name of the server on which this part is being downloaded
     * @conn_id:          The connection id which is used to download this part
     * @collection_name:  The name of the collection from which this part is
     * @subject:          The name of the subject from which this part is
     * @part_num:         The number of the part
     *
     * The download of a part has just started
     */
    signals[PART_DOWNLOAD_START_SIGNAL] = create_signal("part_download_start",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    /**
     * NntpgrabCore::part-done:
     * @obj:                An instance of the NntpgrabCore
     * @servername:         The name of the server on which this part has just been downloaded
     * @conn_id:            The connection id which was used to download this part
     * @collection_name:    The name of the collection from which this part is
     * @subject:            The name of the subject from which this part is
     * @part_num:           The number of the part
     * @size:               The size of the part in bytes. This can be different from the size which was mentioned earlier in other calls like nntpgrab_core_schedular_foreach_task()
     *
     * The download of a part has completed successfully
     */
    signals[PART_DONE_SIGNAL] =           create_signal("part_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    /**
     * NntpgrabCore::part-failed:
     * @obj:                An instance of the NntpgrabCore
     * @servername:         The name of the server on which this part was tried
     * @conn_id:            The connection id which was used to try to download this part
     * @collection_name:    The name of the collection from which this part is
     * @subject:            The name of the subject from which this part is
     * @part_num:           The number of the part
     * @size:               The size of the part in bytes
     * @all_servers_tried:  If TRUE, all configured servers have tried to download this file and none of them succeeded
     *
     * The download of a part has failed (probably because the part isn't available on the server)
     */
    signals[PART_FAILED_SIGNAL] =         create_signal("part_failed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT_BOOLEAN,
                                                        G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_BOOLEAN);

    /**
     * NntpgrabCore::part-progress-update:
     * @obj:                An instance of the NntpgrabCore
     * @servername:         The name of the server on which this part is being downloaded
     * @conn_id:            The connection id which is being used to download this part
     * @collection_name:    The name of the collection from which this part is
     * @subject:            The name of the subject from which this part is
     * @part_num:           The number of the part
     * @bytes_downloaded:   The number of bytes already downloaded from this part
     * @bytes_total:        The total size of this part in bytes. This can be different from the size which was mentioned earlier in other calls like nntpgrab_core_schedular_foreach_task()
     *
     * Every 1/10th second of a part download, this message will be sent
     */
    signals[PART_PROGRESS_UPDATE_SIGNAL] = create_signal("part_progress_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT_INT,
                                                         G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    /**
     * NntpgrabCore::traffic-monitor-update:
     * @obj:                An instance of the NntpgrabCore
     * @bytes_received1:    The number of bytes received in now() - 10
     * @bytes_received2:    The number of bytes received in now() - 9
     * @bytes_received3:    The number of bytes received in now() - 8
     * @bytes_received4:    The number of bytes received in now() - 7
     * @bytes_received5:    The number of bytes received in now() - 6
     * @bytes_received6:    The number of bytes received in now() - 5
     * @bytes_received7:    The number of bytes received in now() - 4
     * @bytes_received8:    The number of bytes received in now() - 3
     * @bytes_received9:    The number of bytes received in now() - 2
     * @bytes_received10:   The number of bytes received in now() - 1
     * @stamp:              The timestamp of the last measurement
     * @average:            The average number of bytes per second
     *
     * Every second, traffic statistics about the last 10 seconds are given
     */
    signals[TRAFFIC_MONITOR_UPDATE_SIGNAL] = create_signal("traffic_monitor_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__INT_INT_INT_INT_INT_INT_INT_INT_INT_INT_INT64_DOUBLE,
                                                         G_TYPE_NONE, 12, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT64, G_TYPE_DOUBLE);

    /**
     * NntpgrabCore::collection-added:
     * @obj:                                An instance of the NntpgrabCore
     * @collection_name:                    The name of the collection which was just added
     * @poster:             (allow-none):   The poster of this collection. If multiple posters are involved, this value will be NULL
     *
     * A new collection was added
     */
    signals[COLLECTION_ADDED_SIGNAL] =    create_signal("collection_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    /**
     * NntpgrabCore::collection-removed:
     * @obj:                An instance of the NntpgrabCore
     * @collection_name:    The name of the collection which was just removed
     *
     * A collection was removed
     */
    signals[COLLECTION_REMOVED_SIGNAL] =  create_signal("collection_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    /**
     * NntpgrabCore::collection-modified:
     * @obj:                                An instance of the NntpgrabCore
     * @collection_name:                    The name of the collection which was just changed
     * @poster:             (allow-none):   The name of the poster of this collection. This value can be NULL if multiple posters are involved
     *
     * The poster of a collection was modified
     */
    signals[COLLECTION_REMOVED_SIGNAL] =  create_signal("collection_modified",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    /**
     * NntpgrabCore::file-added:
     * @obj:                                            An instance of the NntpgrabCore
     * @collection_name:                                The name of the collection in which this file belongs
     * @subject:                                        The subject of the file which was just added
     * @poster:                                         The poster of the file which was just added
     * @stamp:                                          The stamp mentioning when the first part of this file was posted
     * @file_size:                                      The size of this file in bytes
     * @total_size:                                     The size of the entire collection in bytes
     * @total_size_remaining:                           The number of bytes which still need to be downloaded from the entire collection
     * @status:                                         The current status of this file. This is of the type #NGTaskState
     * @num_parts:                                      The number of parts of which this file consists
     * @groups:                 (element-type utf8):    A list containing the groups in which this file is posted
     *
     * A file was added to a already existing collection
     */
    signals[FILE_ADDED_SIGNAL] =          create_signal("file_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_UINT64_UINT64_UINT64_UINT64_INT_INT_POINTER,
                                                        G_TYPE_NONE, 10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);

    /**
     * NntpgrabCore::file-removed:
     * @obj:                    An instance of the NntpgrabCore
     * @collection_name:        The name of the collection in which this file was situated
     * @subject:                The subject of the file
     * @total_size:             The new total size of the entire collection
     * @total_size_remaining:   The new total size remaining of the entire collection
     *
     * A file was removed from the download queue
     */
    signals[FILE_REMOVED_SIGNAL] =        create_signal("file_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_UINT64_UINT64,
                                                        G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64);

    /**
     * NntpgrabCore::file-download-state-update:
     * @obj:                    An instance of the NntpgrabCore
     * @collection_name:        The collection name
     * @subject:                The subject
     * @num_parts_total:        The number of parts of which this file consists
     * @num_parts_done:         The number of parts of this file which are already downloaded
     * @num_parts_failed:       The number of parts of this file which failed to download
     * @file_size:              The size of this file in bytes
     * @file_size_remaining:    The number of bytes remaining from this file
     * @total_size:             The total size of the entire collection in bytes
     * @total_size_remaining:   The number of bytes remaining from the entire collection
     *
     * The state of a file in the download queue has changed
     */
    signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL] =  create_signal("file_download_state_update",
                                                                G_OBJECT_CLASS_TYPE (klass),
                                                                G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                                0,
                                                                NULL, NULL,
                                                                nntpgrab_marshal_VOID__STRING_STRING_INT_INT_INT_UINT64_UINT64_UINT64_UINT64,
                                                                G_TYPE_NONE, 9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64);

    /**
     * NntpgrabCore::file-state-changed:
     * @obj:                                An instance of the NntpgrabCore
     * @collection_name:                    The name of the collection in which the state changed
     * @subject:                            The subject of the file which changed state
     * @real_filename:      (allow-none):   The file name of the physical resulting file. Will be NULL if the file isn't completely downloaded and decoded yet
     * @old_state:                          The old state of the file
     * @new_state:                          The new state of the file
     *
     * A file in the download queue was restarted
     */
    signals[FILE_STATE_CHANGED_SIGNAL] =  create_signal("file_state_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    /**
     * NntpgrabCore::connection-connecting:
     * @obj:                An instance of the NntpgrabCore
     * @servername:         The name of the server to which a connection is being made
     * @conn_id:            An identifier for this connection
     *
     * A new connection is being made to a usenet server
     */
    signals[CONNECTION_CONNECTING_SIGNAL] =    create_signal("connection_connecting",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT,
                                                             G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

    /**
     * NntpgrabCore::connection-connected:
     * @obj:                An instance of the NntpgrabCore
     * @servername:         The name of the server to which a connection has just been made
     * @conn_id:            An identifier for this connection
     * @welcome_msg:        The welcome message which was returned from the server
     *
     * A connection attempt has succeeded
     */
    signals[CONNECTION_CONNECTED_SIGNAL] =     create_signal("connection_connected",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                             G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    /**
     * NntpgrabCore::connection-disconnect:
     * @obj:                             An instance of the NntpgrabCore
     * @servername:                      The name of the server which was just disconnected
     * @conn_id:                         An identifier for this connection
     * @disconnect_type:                 One of the values from #NNTPDisconnectType
     * @reason:           (allow-none):  The reason which indicates why the disconnect occured. Can be NULL
     *
     * A connection to a usenet server was disconnected
     */
    signals[CONNECTION_DISCONNECT_SIGNAL] = create_signal("connection_disconnect",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_ENUM_STRING,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);

    /**
     * NntpgrabCore::schedular-state-changed:
     * @obj:                        An instance of the NntpgrabCore
     * @new_state:                  The new state of the schedular
     * @reason:     (allow-none):   The reason why the schedular changed state. Will be NULL if it's caused by request of the user (nntpgrab_core_schedular_start() / nntpgrab_core_schedular_stop())
     *
     * The state of the schedular has changed
     */
    signals[SCHEDULAR_STATE_CHANGED_SIGNAL] = create_signal("schedular_state_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__INT_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);


    /**
     * NntpgrabCore::log-message:
     * @obj:                    An instance of the NntpgrabCore
     * @component:              The component which emit this log message
     * @log_level:              One of the values from #NGLogLevel
     * @msg:                    The log message which was emit
     *
     * A log message has been emitted
     */
    signals[LOG_MESSAGE_SIGNAL] =           create_signal("log_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    /**
     * NntpgrabCore::task-moved:
     * @obj:                    An instance of the NntpgrabCore
     * @orig_collection_name:   The name of the original collection of which the file was a member of
     * @subject:                The subject of the file
     * @new_collection_name:    The name of the new collection to which the file is moved (which can be the same of @orig_collection_name)
     * @old_position:           The original position of the file in the original collection
     * @new_position:           The new position of the file in the new collection
     *
     * The position of a task in the download queue has changed
     */
    signals[TASK_MOVED_SIGNAL] =            create_signal("task_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                          G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    /**
     * NntpgrabCore::collection-moved:
     * @obj:                    An instance of the NntpgrabCore
     * @collection_name:        The name of the collection which has just been moved
     * @old_position:           The original position in the download queue of the collection
     * @new_position:           The new position in the download queue of the collection
     *
     * The position of a collection in the download queue has changed
     */
    signals[COLLECTION_MOVED_SIGNAL] =      create_signal("collection_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_INT,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    /**
     * NntpgrabCore::collection-downloaded:
     * @obj:                     An instance of the NntpgrabCore
     * @collection_name:         The name of the collection which has been fully downloaded
     *
     * A collection has been completely downloaded
     */
    signals[COLLECTION_DOWNLOADED_SIGNAL] =   create_signal("collection_downloaded",
                                                            G_OBJECT_CLASS_TYPE (klass),
                                                            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                            0,
                                                            NULL, NULL,
                                                            nntpgrab_marshal_VOID__STRING,
                                                            G_TYPE_NONE, 1, G_TYPE_STRING);

    /**
     * NntpgrabCore::all-downloads-completed:
     * @obj:                   An instance of the NntpgrabCore
     *
     * This message is emit when all the items in the download queue are completed
     */
    signals[ALL_DOWNLOADS_COMPLETED_SIGNAL] = create_signal("all_downloads_completed",
                                                            G_OBJECT_CLASS_TYPE (klass),
                                                            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                            0,
                                                            NULL, NULL,
                                                            nntpgrab_marshal_VOID__VOID,
                                                            G_TYPE_NONE, 0);

    /**
     * NntpgrabCore::plugin-loaded:
     * @obj:                   An instance of the NntpgrabCore
     * @plugin_name:           The name of the plugin
     * @is_persistent:         Unused in NNTPGrab 0.7
     *
     * A plugin has been loaded
     */
    signals[PLUGIN_LOADED_SIGNAL] =         create_signal("plugin_loaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_BOOL,
                                                          G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    /**
     * NntpgrabCore::plugin-unloaded:
     * @obj:                 An instance of the NntpgrabCore
     * @plugin_name:         The name of the plugin
     *
     * A plugin has been unloaded
     */
    signals[PLUGIN_UNLOADED_SIGNAL] =       create_signal("plugin_unloaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    /**
     * NntpgrabCore::plugin-event:
     * @obj:                 An instance of the NntpgrabCore
     * @plugin_name:         The name of the plugin
     * @event_name:          The name of the plugin event
     * @values:              The values belonging to this event. This is a NULL-terminated array
     *
     * A plugin has emit an event
     */
    signals[PLUGIN_EVENT_SIGNAL] =          create_signal("plugin_event",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_BOXED,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRV);

    /**
     * NNTPGrab::quit-requested:
     * @obj:                An instance of the NntpgrabCore
     *
     * A request has been received that the frontend needs to quit
     * Frontends need to respond to this event by performing a proper cleanup
     */
    signals[QUIT_REQUESTED_SIGNAL] =        create_signal("quit_requested",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__VOID,
                                                          G_TYPE_NONE, 0);
}

NntpgrabCore *
get_core(void)
{
    if (core_global && core_global->is_finalizing) {
        return NULL;
    }

    return core_global;
}

static gboolean
on_config_changed(Configuration *config, gpointer data)
{
    nntpgrab_core_emit_config_changed(FALSE);
    return FALSE;
}

/***************************************************************/
/*                                                             */
/*                    Public methods                           */
/*                                                             */
/***************************************************************/

NntpgrabCore *
nntpgrab_core_new(void)
{
    return g_object_new((GType) NNTPGRAB_TYPE_CORE, NULL);
}

ngboolean
nntpgrab_core_init(NntpgrabCore *obj, int version, char **err, char **warnings)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
#ifdef WIN32
    char *installation_path;
#endif

    g_return_val_if_fail(core != NULL, FALSE);
    core_global = core;

    if (core->is_initialized) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core already initialized"));
        }
        return FALSE;
    }

    // Check if the API version matches
    if (version != NNTPGRAB_API_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core API mismatch (Core API version = %i, frontend's version = %i)"), NNTPGRAB_API_VERSION, version);
        }
        return FALSE;
    }

#ifdef WIN32
    // Detect the installation directory on Win32
    installation_path = g_win32_get_package_installation_directory_of_module(NULL);
    g_return_val_if_fail(installation_path != NULL, FALSE);

    if (!g_getenv("NNTPGRAB_LIBDIR")) {
        char *path = g_strdup_printf("%s\\lib\\nntpgrab", installation_path);
        g_setenv("NNTPGRAB_LIBDIR", path, TRUE);
        g_free(path);
    }

    if (!g_getenv("NNTPGRAB_WWWDIR")) {
        char *path = g_strdup_printf("%s\\webinterface", installation_path);
        g_setenv("NNTPGRAB_WWWDIR", path, TRUE);
        g_free(path);
    }

    g_free(installation_path);
#endif

    if (g_getenv("NNTPGRAB_DEBUG")) {
        plugins_print_logging();
    }

    if (!plugins_populate_list(err)) {
        return FALSE;
    }

    if (!(core->config = configuration_new(err, warnings))) {
        return FALSE;
    }

    g_signal_connect(core->config, "config_changed", G_CALLBACK(on_config_changed), NULL);

    nntpconnection_init_libproxy();

    if (!(download_queue_init(core->config, err))) {
        return FALSE;
    }

    if (!plugins_load(core, core->config, err)) {
        return FALSE;
    }

    /* Try to load the 'Unpack' and 'Auto-import' plugins if they're available */
    plugins_load_plugin_by_name("Unpack", NULL);
    plugins_load_plugin_by_name("Auto-import", NULL);

    decoder_thread_initialize(core->config);

#if 0
    if (!par2_initialize(err)) {
        return NULL;
    }

    if (!unpack_initialize(err)) {
        return NULL;
    }
#endif

    core->is_initialized = TRUE;

    return TRUE;
}

void
nntpgrab_core_destroy(NntpgrabCore *obj)
{
    g_object_unref(obj);
}

NGList *
nntpgrab_core_config_get_avail_servers(NntpgrabCore *obj)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, NULL);
    g_return_val_if_fail(core->config, NULL);

    return (NGList*) configuration_get_avail_servers(core->config);
}

void
nntpgrab_core_config_free_avail_servers(NntpgrabCore *obj, NGList *servers)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_if_fail(core != NULL);
    g_return_if_fail(core->is_initialized == TRUE);
    g_return_if_fail(core->config);

    configuration_free_avail_servers(core->config, (GList*) servers);
}

ngboolean
nntpgrab_core_config_get_server_info(NntpgrabCore *obj, const char *servername, NGConfigServer *ret)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    NGConfigServer *config;

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(core->config, FALSE);

    config = configuration_get_server_info(core->config, servername);
    if (config) {
        memcpy(ret, config, sizeof(NGConfigServer));
        g_slice_free(NGConfigServer, config);
        return TRUE;
    } else {
        return FALSE;
    }
}

ngboolean
nntpgrab_core_config_add_server(NntpgrabCore *obj, NGConfigServer new_server, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(core->config, FALSE);

    retval = configuration_add_server(core->config, new_server, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_ERROR, "%s", errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

ngboolean
nntpgrab_core_config_del_server(NntpgrabCore *obj, const char *servername, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(core->config, FALSE);

    retval = configuration_del_server(core->config, servername, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_ERROR, "%s", errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

ngboolean
nntpgrab_core_config_edit_server(NntpgrabCore *obj, const char *servername, NGConfigServer server, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(core->config, FALSE);

    retval = configuration_edit_server(core->config, servername, server, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_ERROR, "%s", errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

NGConfigOpts
nntpgrab_core_config_get_opts(NntpgrabCore *obj)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    NGConfigOpts dummy;

    memset(&dummy, 0, sizeof(dummy));
    g_return_val_if_fail(core != NULL, dummy);
    g_return_val_if_fail(core->is_initialized == TRUE, dummy);
    g_return_val_if_fail(core->config, dummy);

    return configuration_get_opts(core->config);
}

void
nntpgrab_core_config_set_opts(NntpgrabCore *obj, NGConfigOpts opts)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);
    char *errmsg = NULL;

    g_return_if_fail(core != NULL);
    g_return_if_fail(core->is_initialized == TRUE);
    g_return_if_fail(core->config);

    configuration_set_opts(core->config, opts);

    if (!configuration_save(core->config, &errmsg)) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_ERROR, "%s", errmsg);
        g_free(errmsg);
    }
}

gboolean
nntpgrab_core_schedular_start(NntpgrabCore *obj)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return download_thread_start(core->config);
}

static gpointer
schedular_stop_func(gpointer data)
{
    NntpgrabCore *core = NNTPGRAB_CORE(data);

    g_return_val_if_fail(core != NULL, NULL);
    g_return_val_if_fail(core->is_initialized == TRUE, NULL);

    download_thread_stop(FALSE, NULL);

    return NULL;
}

ngboolean
nntpgrab_core_schedular_stop(NntpgrabCore *obj, ngboolean wait)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    if (wait) {
        return download_thread_stop(FALSE, NULL);
    } else {
        // Is the schedular queue already stopped?
        if (download_thread_get_state() != SCHEDULAR_STATE_RUNNING) {
            return FALSE;
        }

        g_thread_create(schedular_stop_func, obj, FALSE, NULL);

        return TRUE;
    }
}

NGSchedularState
nntpgrab_core_schedular_get_state(NntpgrabCore *obj)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, SCHEDULAR_STATE_STOPPED);
    g_return_val_if_fail(core->is_initialized == TRUE, SCHEDULAR_STATE_STOPPED);

    return (download_thread_get_state());
}

ngboolean
nntpgrab_core_schedular_add_task_to_queue(NntpgrabCore *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, nguint64 file_size, NGList *groups, NGList *parts, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return download_queue_add_file_to_queue(collection_name, subject, poster, stamp, file_size, groups, parts, errmsg);
}

ngboolean
nntpgrab_core_schedular_del_task_from_queue(NntpgrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    /* NOTE: subject MIGHT be NULL ! */

    return download_queue_del_file_from_queue(collection_name, subject, errmsg);
}

ngboolean
nntpgrab_core_schedular_restart_task(NntpgrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    /* NOTE: subject MIGHT be NULL ! */

    return download_queue_restart_file(collection_name, subject, errmsg);
}

ngboolean
nntpgrab_core_schedular_save_queue(NntpgrabCore *obj, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return download_queue_save(errmsg);
}

void
nntpgrab_core_schedular_foreach_task(NntpgrabCore *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, void *data)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_if_fail(core != NULL);
    g_return_if_fail(core->is_initialized == TRUE);

    download_queue_foreach_file(collection_func, file_func, group_func, data);
}

ngboolean
nntpgrab_core_schedular_move_task(NntpgrabCore *obj, const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(collection_name_src != NULL, FALSE);
    g_return_val_if_fail(subject_src != NULL, FALSE);
    g_return_val_if_fail(collection_name_dest != NULL, FALSE);
    g_return_val_if_fail(position_dest >= -1, FALSE);

    return download_queue_move_file(collection_name_src, subject_src, collection_name_dest, position_dest);
}

ngboolean
nntpgrab_core_schedular_move_collection(NntpgrabCore *obj, const char *collection_name, int new_position)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(new_position >= -1, FALSE);

    return download_queue_move_collection(collection_name, new_position);
}

ngboolean
nntpgrab_core_schedular_mark_task_optional(NntpgrabCore *obj, const char *collection_name, const char *subject, ngboolean is_optional)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(subject != NULL, FALSE);

    return download_queue_mark_task_optional(collection_name, subject, is_optional);
}

NGList *
nntpgrab_core_plugins_get_avail_plugins(NntpgrabCore *obj)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, NULL);
    g_return_val_if_fail(core->is_initialized == TRUE, NULL);

    return plugins_get_avail_plugins();
}

void
nntpgrab_core_plugins_free_avail_plugins(NntpgrabCore *obj, NGList *plugins)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_if_fail(core != NULL);
    g_return_if_fail(core->is_initialized == TRUE);

    plugins_free_avail_plugins(plugins);
}

ngboolean
nntpgrab_core_plugins_get_plugin_info(NntpgrabCore *obj, const char *plugin_name, NNTPGrabPluginInfo *plugin_info)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return plugins_get_plugin_info(plugin_name, plugin_info);
}

ngboolean
nntpgrab_core_plugins_load_plugin(NntpgrabCore *obj, const char *plugin_name, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return plugins_load_plugin_by_name(plugin_name, errmsg);
}

ngboolean
nntpgrab_core_plugins_unload_plugin(NntpgrabCore *obj, const char *plugin_name, char **errmsg)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return plugins_unload_plugin_by_name(plugin_name, errmsg);
}

ngboolean
nntpgrab_core_plugins_set_persistent(NntpgrabCore *obj, const char *plugin_name, ngboolean persistent)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    return plugins_set_persistent(plugin_name, persistent);
}

ngboolean
nntpgrab_core_embedded_server_start(NntpgrabCore *obj, int port, char **errmsg)
{
    ngboolean retval = FALSE;
    NNTPGrabPluginInfo plugin_info;
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(core->is_initialized == TRUE, FALSE);

    if (!plugins_get_plugin_info("JSON-RPC", &plugin_info)) {
        *errmsg = g_strdup_printf(_("Unable to start the embedded server as the JSON-RPC plugin could not be found"));
        return FALSE;
    }

    if (!plugin_info.is_loaded &&
        !plugins_load_plugin_by_name("JSON-RPC", errmsg)) {

        return FALSE;
    }

    ng_plugin_call(NULL, "webserver_start", port, errmsg, &retval);
    ng_plugin_call(NULL, "webserver_ignore_config_changes");

    return retval;
}

void
nntpgrab_core_set_emit_log_messages(NntpgrabCore *obj, ngboolean val)
{
    NntpgrabCore *core = NNTPGRAB_CORE(obj);

    g_return_if_fail(core != NULL);
    g_return_if_fail(core->is_initialized == TRUE);

    plugins_set_emit_log_messages_to_frontend(val);
}
