/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <glib-object.h>

#include "nntpgrab_plugin.h"
#include "nntpgrab_utils.h"

#include "json.h"
#include "jsonrpc.h"

#undef max
#define max(a,b) (((a) > (b)) ? (a) : (b))

void jsonrpc_tcp_emit_event(const char *json_data);

static void
emit_event(const char *event_name, struct json_object *event)
{
    struct json_object *tcp_event = json_object_new_object();

    jsonrpc_add_event((char*) event_name, json_object_get(event));

    json_object_object_add(tcp_event, "id", NULL);
    json_object_object_add(tcp_event, "method", json_object_new_string((char*) event_name));
    json_object_object_add(tcp_event, "params", event);
    jsonrpc_tcp_emit_event(json_object_to_json_string(tcp_event));

    json_object_put(tcp_event);
}

static void
config_changed (NGPlugin *plugin_data, gpointer data)
{
    struct json_object *event = json_object_new_object();

    emit_event("config_changed", event);

    json_object_put(event);
}

static void
part_download_start (NGPlugin *plugin_data, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "part_num", json_object_new_int(part_num));

    emit_event("part_download_start", event);

    json_object_put(event);
}

static void
part_done (NGPlugin *plugin_data, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "part_num", json_object_new_int(part_num));
    json_object_object_add(event, "size", json_object_new_int(size));

    emit_event("part_done", event);

    json_object_put(event);
}

static void
part_failed (NGPlugin *plugin_data, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "part_num", json_object_new_int(part_num));
    json_object_object_add(event, "size", json_object_new_int(size));
    json_object_object_add(event, "all_servers_tried", json_object_new_boolean(all_servers_tried));

    emit_event("part_failed", event);

    json_object_put(event);
}

static void
traffic_monitor_update (NGPlugin *plugin_data, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 stamp, double average, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "bytes_received1", json_object_new_int(bytes_received1));
    json_object_object_add(event, "bytes_received2", json_object_new_int(bytes_received2));
    json_object_object_add(event, "bytes_received3", json_object_new_int(bytes_received3));
    json_object_object_add(event, "bytes_received4", json_object_new_int(bytes_received4));
    json_object_object_add(event, "bytes_received5", json_object_new_int(bytes_received5));
    json_object_object_add(event, "bytes_received6", json_object_new_int(bytes_received6));
    json_object_object_add(event, "bytes_received7", json_object_new_int(bytes_received7));
    json_object_object_add(event, "bytes_received8", json_object_new_int(bytes_received8));
    json_object_object_add(event, "bytes_received9", json_object_new_int(bytes_received9));
    json_object_object_add(event, "bytes_received10", json_object_new_int(bytes_received10));
    json_object_object_add(event, "stamp", json_object_new_int(stamp));
    json_object_object_add(event, "average", json_object_new_double(average));

    emit_event("traffic_monitor_update", event);

    json_object_put(event);
}

static void
part_progress_update (NGPlugin *plugin_data, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "part_num", json_object_new_int(part_num));
    json_object_object_add(event, "bytes_downloaded", json_object_new_int(bytes_downloaded));
    json_object_object_add(event, "bytes_total", json_object_new_int(bytes_total));

    emit_event("part_progress_update", event);

    json_object_put(event);
}

static void
collection_added (NGPlugin *plugin_data, const char *collection_name, const char *poster, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "poster", json_object_new_string((char*) poster));

    emit_event("collection_added", event);

    json_object_put(event);
}

static void
collection_removed (NGPlugin *plugin_data, const char *collection_name, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));

    emit_event("collection_removed", event);

    json_object_put(event);
}

static void
collection_modified (NGPlugin *plugin_data, const char *collection_name, const char *poster, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "poster", json_object_new_string((char*) poster));

    emit_event("collection_modified", event);

    json_object_put(event);
}

static void
file_added (NGPlugin *plugin_data, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_size, guint64 total_size_remaining, NGTaskState status, int num_parts, GList *groups, gpointer data)
{
    struct json_object *event = json_object_new_object();
    struct json_object *obj_groups = json_object_new_array();
    GList *list;
    int i_file_size = 0;
    int i_total_size = 0;
    int i_total_size_remaining = 0;

    if (file_size != 0) {
        i_file_size = max(1, file_size / 1024);
    }
    if (total_size != 0) {
        i_total_size = max(1, total_size / 1024);
    }
    if (total_size_remaining != 0) {
        i_total_size_remaining = max(1, total_size_remaining / 1024);
    }

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "poster", json_object_new_string((char*) poster));
    json_object_object_add(event, "stamp", json_object_new_int(stamp));
    json_object_object_add(event, "file_size", json_object_new_int(i_file_size));
    json_object_object_add(event, "total_size", json_object_new_int(i_total_size));
    json_object_object_add(event, "total_size_remaining", json_object_new_int(i_total_size_remaining));
    json_object_object_add(event, "status", json_object_new_int(status));
    json_object_object_add(event, "num_parts", json_object_new_int(num_parts));
    json_object_object_add(event, "groups", obj_groups);

    list = groups;
    while (list) {
        char *groupname = list->data;

        json_object_array_add(obj_groups, json_object_new_string(groupname));

        list = g_list_next(list);
    }

    emit_event("file_added", event);

    json_object_put(event);
}

static void
file_removed (NGPlugin *plugin_data, const char *collection_name, const char *subject, guint64 total_size, guint64 total_size_remaining, gpointer data)
{
    struct json_object *event = json_object_new_object();
    int i_total_size = 0;
    int i_total_size_remaining = 0;

    if (total_size != 0) {
        i_total_size = max(1, total_size / 1024);
    }
    if (total_size_remaining != 0) {
        i_total_size_remaining = max(1, total_size_remaining / 1024);
    }

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "total_size", json_object_new_int(i_total_size));
    json_object_object_add(event, "total_size_remaining", json_object_new_int(i_total_size_remaining));

    emit_event("file_removed", event);

    json_object_put(event);
}

static void
file_download_state_update (NGPlugin *plugin_data, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_size, guint64 total_size_remaining, gpointer data)
{
    struct json_object *event = json_object_new_object();
    int i_file_size = 0;
    int i_file_size_remaining = 0;
    int i_total_size = 0;
    int i_total_size_remaining = 0;

    if (file_size != 0) {
        i_file_size = max(1, file_size / 1024);
    }
    if (file_size_remaining) {
        i_file_size_remaining = max(1, file_size_remaining / 1024);
    }
    if (total_size != 0) {
        i_total_size = max(1, total_size / 1024);
    }
    if (total_size_remaining != 0) {
        i_total_size_remaining = max(1, total_size_remaining / 1024);
    }

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "num_parts_total", json_object_new_int(num_parts_total));
    json_object_object_add(event, "num_parts_done", json_object_new_int(num_parts_done));
    json_object_object_add(event, "num_parts_failed", json_object_new_int(num_parts_failed));
    json_object_object_add(event, "file_size", json_object_new_int(i_file_size));
    json_object_object_add(event, "file_size_remaining", json_object_new_int(i_file_size_remaining));
    json_object_object_add(event, "total_size", json_object_new_int(i_total_size));
    json_object_object_add(event, "total_size_remaining", json_object_new_int(i_total_size_remaining));

    emit_event("file_download_state_update", event);

    json_object_put(event);
}

static void
file_state_changed (NGPlugin *plugin_data, const char *collection_name, const char *subject, const char *real_filename, NGTaskState old_state, NGTaskState new_state, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "real_filename", json_object_new_string((char*) (real_filename ? real_filename : "")));
    json_object_object_add(event, "old_state", json_object_new_int(old_state));
    json_object_object_add(event, "new_state", json_object_new_int(new_state));

    emit_event("file_state_changed", event);

    json_object_put(event);
}

static void
connection_connecting (NGPlugin *plugin_data, const char *servername, int conn_id, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));

    emit_event("connection_connecting", event);

    json_object_put(event);
}

static void
connection_connected (NGPlugin *plugin_data, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "welcome_msg", json_object_new_string((char*) welcome_msg));

    emit_event("connection_connected", event);

    json_object_put(event);
}

static void
connection_disconnect (NGPlugin *plugin_data, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *msg, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "servername", json_object_new_string((char*) servername));
    json_object_object_add(event, "conn_id", json_object_new_int(conn_id));
    json_object_object_add(event, "disconnect_type", json_object_new_int(disconnect_type));
    json_object_object_add(event, "reason", json_object_new_string((char*) (msg ? msg : "")));

    emit_event("connection_disconnect", event);

    json_object_put(event);
}

static void
schedular_state_changed (NGPlugin *plugin_data, NGSchedularState new_state, const char *reason, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "new_state", json_object_new_int(new_state));
    json_object_object_add(event, "reason", json_object_new_string((char*) (reason ? reason : "")));

    emit_event("schedular_state_changed", event);

    json_object_put(event);
}

static void
log_message (NGPlugin *plugin_data, const char *component, NGLogLevel log_level, const char *msg, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "component", json_object_new_string((char*) component));
    json_object_object_add(event, "log_level", json_object_new_int(log_level));
    json_object_object_add(event, "msg", json_object_new_string((char*) msg));

    emit_event("log_message", event);

    json_object_put(event);
}

static void
task_moved (NGPlugin *plugin_data, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "orig_collection_name", json_object_new_string((char*) orig_collection_name));
    json_object_object_add(event, "subject", json_object_new_string((char*) subject));
    json_object_object_add(event, "new_collection_name", json_object_new_string((char*) new_collection_name));
    json_object_object_add(event, "old_position", json_object_new_int(old_position));
    json_object_object_add(event, "new_position", json_object_new_int(new_position));

    emit_event("task_moved", event);

    json_object_put(event);
}

static void
collection_moved (NGPlugin *plugin_data, const char *collection_name, int old_position, int new_position, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "collection_name", json_object_new_string((char*) collection_name));
    json_object_object_add(event, "old_position", json_object_new_int(old_position));
    json_object_object_add(event, "new_position", json_object_new_int(new_position));

    emit_event("collection_moved", event);

    json_object_put(event);
}

static void
all_downloads_completed (NGPlugin *plugin_data, gpointer data)
{
    struct json_object *event = json_object_new_object();

    emit_event("all_downloads_completed", event);

    json_object_put(event);
}

static void
plugin_loaded (NGPlugin *plugin_data, const char *plugin_name, gboolean is_persistent, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "plugin_name", json_object_new_string((char*) plugin_name));
    json_object_object_add(event, "is_persistent", json_object_new_boolean(is_persistent));

    emit_event("plugin_loaded", event);

    json_object_put(event);
}

static void
plugin_unloaded (NGPlugin *plugin_data, const char *plugin_name, gpointer data)
{
    struct json_object *event = json_object_new_object();

    json_object_object_add(event, "plugin_name", json_object_new_string((char*) plugin_name));

    emit_event("plugin_unloaded", event);

    json_object_put(event);
}

static void
plugin_event (NGPlugin *plugin_data, const char *plugin_name, const char *event_name, const char **values)
{
    struct json_object *event = json_object_new_object();
    struct json_object *obj_values = json_object_new_array();
    int i;

    json_object_object_add(event, "plugin_name", json_object_new_string((char*) plugin_name));
    json_object_object_add(event, "event_name", json_object_new_string((char*) event_name));
    json_object_object_add(event, "values", obj_values);

    i = 0;
    while (values[i]) {
        json_object_array_add(obj_values, json_object_new_string((char*) values[i]));
        i++;
    }

    emit_event("plugin_event", event);

    json_object_put(event);
}

void
jsonrpc_connect_signal_handlers(NGPlugin *plugin_data)
{
    gpointer data = NULL;

    ng_plugin_connect_event(plugin_data, "config_changed", NG_PLUGIN_FUNCTION(config_changed), data);
    ng_plugin_connect_event(plugin_data, "part_download_start", NG_PLUGIN_FUNCTION(part_download_start), data);
    ng_plugin_connect_event(plugin_data, "part_done", NG_PLUGIN_FUNCTION(part_done), data);
    ng_plugin_connect_event(plugin_data, "part_failed", NG_PLUGIN_FUNCTION(part_failed), data);
    ng_plugin_connect_event(plugin_data, "traffic_monitor_update", NG_PLUGIN_FUNCTION(traffic_monitor_update), data);
    ng_plugin_connect_event(plugin_data, "part_progress_update", NG_PLUGIN_FUNCTION(part_progress_update), data);
    ng_plugin_connect_event(plugin_data, "collection_added", NG_PLUGIN_FUNCTION(collection_added), data);
    ng_plugin_connect_event(plugin_data, "collection_removed", NG_PLUGIN_FUNCTION(collection_removed), data);
    ng_plugin_connect_event(plugin_data, "collection_modified", NG_PLUGIN_FUNCTION(collection_modified), data);
    ng_plugin_connect_event(plugin_data, "file_added", NG_PLUGIN_FUNCTION(file_added), data);
    ng_plugin_connect_event(plugin_data, "file_removed", NG_PLUGIN_FUNCTION(file_removed), data);
    ng_plugin_connect_event(plugin_data, "file_download_state_update", NG_PLUGIN_FUNCTION(file_download_state_update), data);
    ng_plugin_connect_event(plugin_data, "file_state_changed", NG_PLUGIN_FUNCTION(file_state_changed), data);
    ng_plugin_connect_event(plugin_data, "connection_connecting", NG_PLUGIN_FUNCTION(connection_connecting), data);
    ng_plugin_connect_event(plugin_data, "connection_connected", NG_PLUGIN_FUNCTION(connection_connected), data);
    ng_plugin_connect_event(plugin_data, "connection_disconnect", NG_PLUGIN_FUNCTION(connection_disconnect), data);
    ng_plugin_connect_event(plugin_data, "schedular_state_changed", NG_PLUGIN_FUNCTION(schedular_state_changed), data);
    ng_plugin_connect_event(plugin_data, "log_message", NG_PLUGIN_FUNCTION(log_message), data);
    ng_plugin_connect_event(plugin_data, "task_moved", NG_PLUGIN_FUNCTION(task_moved), data);
    ng_plugin_connect_event(plugin_data, "collection_moved", NG_PLUGIN_FUNCTION(collection_moved), data);
    ng_plugin_connect_event(plugin_data, "all_downloads_completed", NG_PLUGIN_FUNCTION(all_downloads_completed), data);
    ng_plugin_connect_event(plugin_data, "plugin_loaded", NG_PLUGIN_FUNCTION(plugin_loaded), data);
    ng_plugin_connect_event(plugin_data, "plugin_unloaded", NG_PLUGIN_FUNCTION(plugin_unloaded), data);
    ng_plugin_connect_event(plugin_data, "plugin_event", NG_PLUGIN_FUNCTION(plugin_event), data);
}

void
jsonrpc_disconnect_signal_handlers(NGPlugin *plugin_data)
{
    gpointer data = NULL;

    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(config_changed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(part_download_start), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(part_done), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(part_failed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(traffic_monitor_update), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(part_progress_update), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(collection_added), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(collection_removed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(collection_modified), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(file_added), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(file_removed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(file_download_state_update), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(file_state_changed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(connection_connecting), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(connection_connected), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(connection_disconnect), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(schedular_state_changed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(log_message), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(task_moved), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(collection_moved), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(all_downloads_completed), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(plugin_loaded), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(plugin_unloaded), data);
    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(plugin_event), data);
}
